## Process the large bam that contains reads from all barcodes and get allele counts
## for the list of SNPs provided in a VCF file
## Generate per cell per chr Allele counts
## Runs viterbi algorithm and output sequence of inferred viterbi state
## Authror: Ruqian Lyu
## Date: 04/11/2020


import os
import docopt
import strutils
import hts
import tables
import sequtils
import private/utils

import math
import streams
import private/graph
import private/findpath

proc sscocaller(threads:int, vcff:string, barcodeFile:string,
                bamfile:string, out_dir:string, mapq:int, 
                minbsq:int, mintotal:int, maxtotal:int, mindp:int, maxdp:int, 
                thetaREF:float, thetaALT:float, cmPmb:float,s_Chrs:seq,barcodeTag:string): int =
  var 
    ibam:Bam
    ivcf:VCF
  if not open(ibam, bamfile, threads=threads, index = true):
      quit "couldn't open input bam"
  if not open(ivcf, vcff, threads=threads):
      quit "couldn't open: vcff"

  var hf = hts.hts_open(cstring(barcodeFile), "r")
  #### TODO : Table size
  var barcodeTable =  newOrderedTable[string,int](initialSize = 1024)
  # var outFileTable = initTable[string,File]()
  var kstr: hts.kstring_t
  kstr.l = 0
  kstr.m = 0
  kstr.s = nil
  var ithSperm=0
  var bulkBam = false
  ## initiate the table with CB as has keys, allele counts (ref object) as elements
  while hts_getline(hf, cint(10), addr kstr) > 0:
    if $kstr.s[0] == "#":
      continue
    var v = $kstr.s
    discard barcodeTable.hasKeyOrPut(v, ithSperm)
    ithSperm+=1 
  if barcodeTable.len == 1 and barcodeTable.hasKey("bulk"):
    echo "running in bulk mode and all reads are regarded as from one sample/cell"
    bulkBam = true
  let initProb:array[stateRef..stateAlt, float]=[0.5,0.5]
  ## iterate through each selected chromosome
  for chrom in s_Chrs:
    # var scTable = initTable[string,SingleSpermRec]()
    var snpIndex = 0
    ## number of non zeros
    var nnsize = 0
    var scSpermSeq:SeqSpermViNodes
    ## matches with the order in barcodeTable
    scSpermSeq.setLen(barcodeTable.len)
    var outFileSNPanno,outFileTotalCountMtx,outFileAltCountMtx,outFileVStateMtx,viSegmentInfo:FileStream

    try:
      outFileSNPanno = openFileStream(out_dir & chrom & "_snpAnnot.txt", fmWrite)
      outFileTotalCountMtx = openFileStream(out_dir & chrom & "_totalCount.mtx", fmWrite)
      outFileAltCountMtx = openFileStream(out_dir & chrom & "_altCount.mtx", fmWrite)
      outFileVStateMtx = openFileStream(out_dir & chrom & "_vi.mtx", fmWrite)
      viSegmentInfo = openFileStream(out_dir & chrom & "_viSegInfo.txt", fmWrite)
      #echo strm.readLine()
    except:
      stderr.write getCurrentExceptionMsg()
    
    ## write headers to those mtx files and the first line place holder for total_row total_column total_entry
    outFileTotalCountMtx.writeLine("%%MatrixMarket matrix coordinate integer general")
    outFileTotalCountMtx.writeLine(' '.repeat(50))
    outFileAltCountMtx.writeLine("%%MatrixMarket matrix coordinate integer general")
    outFileAltCountMtx.writeLine(' '.repeat(50))
    outFileVStateMtx.writeLine("%%MatrixMarket matrix coordinate integer general")
    outFileVStateMtx.writeLine(' '.repeat(50))
    outFileSNPanno.writeLine("POS" & "\t" & "REF" & "\t" & "ALT" )

    for rec in ivcf.query(chrom):
      if rec.ALT.len > 1: continue
      if rec.ALT.len == 0: continue
      ## alleleCountTable contains for this SNP.POS, each cell barcode's allele counts 
      var alleleCountTable = countAllele(rec=rec, ibam=ibam, chrom=chrom, mapq=mapq, barcodeTable=barcodeTable,minbsq=minbsq,
                                        maxTotalReads = maxtotal, minTotalReads = mintotal,bulkBam = bulkBam, barcodeTag = barcodeTag)
      if alleleCountTable.len==0: continue

      var rec_alt:char
      rec_alt = rec.ALT[0][0]
      # out_line = $rec.CHROM & "\t" & $rec.POS & "\t" & rec.REF[0] & "\t" & rec_alt
      ## add to snpAnnoSeq, later write to SNPannot file, which contains SNP.pos, SNP.ref,SNP.alt; The rowAnnotations
      snpIndex += 1
      outFileSNPanno.writeLine($rec.POS & "\t" & $rec.REF[0] & "\t" & $rec_alt )
      discard addViNode(barcodeTable = barcodeTable,  alleleCountTable = alleleCountTable,  scSpermSeq = scSpermSeq,
               outFileTotalCountMtx = outFileTotalCountMtx, outFileAltCountMtx  = outFileAltCountMtx, nnsize = nnsize,
               mindp = mindp, maxdp = maxdp, thetaRef = thetaRef, thetaAlt = thetaAlt, snpIndex = snpIndex, rec=rec,
               initProb = initProb, cmPmb = cmPmb)
    var posEnd: int64
    var inferProb,reverseProb = 0.0
    var currentEm: array[stateRef..stateAlt, float]
    var lastNode: ViNode
    var spermVNseq: SpermViNodes
    var ithSNP: int
    ## trans/no trans

    for ithSperm in 0..(scSpermSeq.len-1):
      ## rightGap,leftGap = [0.0,0.0]
      spermVNseq = scSpermSeq[ithSperm]
      if spermVNseq.viNodeseq.len==0: continue
      lastNode = spermVNseq.viNodeseq[high(spermVNseq.viNodeseq)]
      currentEm = getEmission(thetaRef=thetaRef,thetaAlt=thetaAlt,cRef=lastNode.cRef, cAlt=lastNode.cAlt)
      if lastNode.pathScore[stateRef] > lastNode.pathScore[stateAlt]:
        scSpermSeq[ithSperm].viNodeseq[high(scSpermSeq[ithSperm].viNodeseq)].state = stateRef
        ithSNP = scSpermSeq[ithSperm].spermSnpIndexLookUp[high(scSpermSeq[ithSperm].viNodeseq)+1]
        outFileVStateMtx.writeLine($ithSNP & " " & $(ithSperm+1) & " 1")
        inferProb  = currentEm[stateRef]
        reverseProb = currentEm[stateAlt]
      else:
        scSpermSeq[ithSperm].viNodeseq[high(scSpermSeq[ithSperm].viNodeseq)].state = stateAlt
        ithSNP = scSpermSeq[ithSperm].spermSnpIndexLookUp[high(scSpermSeq[ithSperm].viNodeseq)+1]
        outFileVStateMtx.writeLine($ithSNP & " " & $(ithSperm+1) & " 2")
        inferProb  = currentEm[stateAlt]
        reverseProb = currentEm[stateRef]
      posEnd = lastNode.pos
      ## call pathTrackBack will write the most probably hidden state seq and viterbi segment info to the 
      ## relevant File Streams.

      discard pathTrackBack(currentSperm = scSpermSeq[ithSperm], ithSperm = ithSperm,  thetaRef = thetaRef, 
                           thetaAlt = thetaAlt, cmPmb = cmPmb,outFileVStateMtx = outFileVStateMtx,
                           viSegmentInfo = viSegmentInfo, posEnd = posEnd,  inferProb = inferProb,reverseProb = reverseProb)

    outFileTotalCountMtx.setPosition(49)
    outFileVStateMtx.setPosition(49)
    outFileAltCountMtx.setPosition(49)
    outFileAltCountMtx.write($snpIndex & " " & $barcodeTable.len & " " & $nnsize)
    outFileVStateMtx.write($snpIndex & " " & $barcodeTable.len & " " & $nnsize)
    outFileTotalCountMtx.write($snpIndex & " " & $barcodeTable.len & " " & $nnsize)
    
    # outHeaderMtx.writeLine() 
    outFileSNPanno.close()
    outFileTotalCountMtx.close()
    outFileAltCountMtx.close()
    # outHeaderMtx.close()
    outFileVStateMtx.close()
    viSegmentInfo.close()
  ibam.close()
  ivcf.close()
  return 0

when(isMainModule):
  let version = "0.2.2"
  var doc = format("""
  $version

  Usage:
      sscocaller [options] <BAM> <VCF> <barcodeFile> <out_prefix>

Arguments:

  <BAM> the read alignment file with records of single-cell DNA reads
  
  <VCF> the variant call file with records of SNPs

  <barcodeFile> the text file containing the list of cell barcodes

  <out_prefix>  the prefix of output files


Options:
  -t --threads <threads>  number of BAM decompression threads [default: 4]
  -cb --cellbarcode <cellbarcode>  the cell barcode tag, by default it is CB
  -MQ --minMAPQ <mapq>  Minimum MAPQ for read filtering [default: 20]
  -BQ --baseq <baseq>  base quality threshold for a base to be used for counting [default: 13]
  -CHR --chrom <chrom>  the selected chromsome (whole genome if not supplied,separate by comma if multiple chroms)
  -minDP --minDP <minDP>  the minimum DP for a SNP to be included in the output file [default: 1]
  -maxDP --maxDP <maxDP>  the maximum DP for a SNP to be included in the output file [default: 5]
  -maxTotalDP --maxTotalDP <maxTotalDP>  the maximum DP across all barcodes for a SNP to be included in the output file [default: 30]
  -minTotalDP --minTotalDP <minTotalDP>  the minimum DP across all barcodes for a SNP to be included in the output file [default: 10]          
  -chrName --chrName  the chr names with chr prefix or not, if not supplied then no prefix
  -thetaREF --thetaREF <thetaREF>  the theta for the binomial distribution conditioning on hidden state being REF [default: 0.1]
  -thetaALT --thetaALT <thetaALT>  the theta for the binomial distribution conditioning on hidden state being ALT [default: 0.9]
  -cmPmb --cmPmb <cmPmb>  the average centiMorgan distances per megabases default 0.1 cm per Mb [default 0.1]
  -h --help  show help

  Examples
      ./sscocaller --threads 10 AAAGTAGCACGTCTCT-1.raw.bam AAAGTAGCACGTCTCT-1.raw.bam.dp3.alt.vcf.gz barcodeFile.tsv ./percell/ccsnp
""" % ["version", version])

  let args = docopt(doc, version=version)
  
  var
    threads:int
    vcff:string
    barcodeFile:string
    bamfile:string
    selectedChrs:string
    out_dir:string
    mapq:int
    minbsq:int
    mintotal:int
    maxtotal:int
    mindp:int
    maxdp:int
    thetaREF:float
    thetaALT:float
    cmPmb:float
    barcodeTag="CB"
  threads = parse_int($args["--threads"])
  barcodeTag = $args["--barcodeTag"]
  mindp = parse_int($args["--minDP"])
  maxdp = parse_int($args["--maxDP"])
  maxtotal = parse_int($args["--maxTotalDP"])
  mintotal = parse_int($args["--minTotalDP"])
  bamfile = $args["<BAM>"]
  barcodeFile = $args["<barcodeFile>"]
  out_dir = $args["<out_prefix>"]
  vcff = $args["<VCF>"]
  mapq = parse_int($args["--minMAPQ"])
  minbsq = parse_int($args["--baseq"])
  thetaRef = parse_float($args["--thetaREF"])
  thetaAlt = parse_float($args["--thetaALT"])
  cmPmb = parse_float($args["--cmPmb"])

  if($args["--chrom"] != "nil"):
    selectedChrs = $args["--chrom"]
  else:
    let a = toSeq(1..19)
    if(not args["--chrName"]):
      let b = map(a, proc (x: int): string = "" & $x)
      let all_Chrs = concat(b,@["X","Y"])
      selectedChrs = all_Chrs.join(",")
    else: 
      let b = map(a, proc (x: int): string = "chr" & $x)
      let all_Chrs = concat(b,@["chrX","chrY"])
      selectedChrs = all_Chrs.join(",")
  let s_Chrs = selectedChrs.split(',')

  discard sscocaller(threads, vcff, barcodeFile, bamfile,
                     out_dir, mapq, minbsq, mintotal, 
                     maxtotal, mindp, maxdp, thetaREF, thetaALT, cmPmb,s_Chrs,barcodeTag)

